@@ -51,15 +51,16 @@ impl<C, E> ReadableCache<E> for C
     }
 }
 
+#[derive(Clone)]
 pub struct NopCache {}
 
 
 #[async_trait]
 impl Cache<()> for NopCache {
-    async fn get<CA: Cacheable + Send + Sync + 'static>(&mut self, cacheable: CA) -> Result<CacheResponse, ()> {
+    async fn get<CA: Cacheable + Send + Sync + 'static>(&mut self, _cacheable: CA) -> Result<CacheResponse, ()> {
         Ok(CacheResponse::Miss)
     }
-    async fn store(&mut self, identity: Vec<u8>) -> Result<(), ()> {
+    async fn store(&mut self, _identity: Vec<u8>) -> Result<(), ()> {
         Ok(())
     }
-}
\ No newline at end of file
+}
