@@ -8,9 +8,10 @@ use rusoto_sqs::Message as SqsMessage;
 use tokio::sync::mpsc::{channel, Receiver, Sender};
 
 use async_trait::async_trait;
-
+use crate::cache::Cache;
 use crate::completion_event_serializer::CompletionEventSerializer;
 use crate::event_emitter::EventEmitter;
+use crate::event_handler::{OutputEvent, Completion};
 
 pub struct CompletionPolicy {
     max_messages: u16,
@@ -37,34 +38,43 @@ impl CompletionPolicy {
     }
 }
 
-pub struct SqsCompletionHandler<CPE, CP, CE, Payload, EE, OA>
+pub struct SqsCompletionHandler<CPE, CP, CE, Payload, EE, OA, CacheT, CacheErr, ProcErr>
     where
         CPE: Debug + Send + Sync + 'static,
         CP: CompletionEventSerializer<CompletedEvent=CE, Output=Payload, Error=CPE> + Send + Sync + 'static,
         Payload: Send + Sync + 'static,
         CE: Send + Sync + Clone + 'static,
         EE: EventEmitter<Event=Payload> + Send + Sync + 'static,
-        OA: Fn(SqsCompletionHandlerActor<CE>, Result<String, String>) + Send + Sync + 'static,
+        OA: Fn(SqsCompletionHandlerActor<CE, ProcErr>, Result<String, String>) + Send + Sync + 'static,
+        CacheT: Cache<CacheErr> + Send + Sync + Clone + 'static,
+        CacheErr: Debug + Clone + Send + Sync + 'static,
+        ProcErr: Debug + Clone + Send + Sync + 'static,
 {
     sqs_client: SqsClient,
     queue_url: String,
     completed_events: Vec<CE>,
+    identities: Vec<Vec<u8>>,
     completed_messages: Vec<SqsMessage>,
     completion_serializer: CP,
     event_emitter: EE,
     completion_policy: CompletionPolicy,
     on_ack: OA,
-    self_actor: Option<SqsCompletionHandlerActor<CE>>
+    self_actor: Option<SqsCompletionHandlerActor<CE, ProcErr>>,
+    cache: CacheT,
+    _p: std::marker::PhantomData<(CacheErr, ProcErr)>
 }
 
-impl<CPE, CP, CE, Payload, EE, OA> SqsCompletionHandler<CPE, CP, CE, Payload, EE, OA>
+impl<CPE, CP, CE, Payload, EE, OA, CacheT, CacheErr, ProcErr> SqsCompletionHandler<CPE, CP, CE, Payload, EE, OA, CacheT, CacheErr, ProcErr>
     where
         CPE: Debug + Send + Sync + 'static,
         CP: CompletionEventSerializer<CompletedEvent=CE, Output=Payload, Error=CPE> + Send + Sync + 'static,
         Payload: Send + Sync + 'static,
         CE: Send + Sync + Clone + 'static,
         EE: EventEmitter<Event=Payload> + Send + Sync + 'static,
-        OA: Fn(SqsCompletionHandlerActor<CE>, Result<String, String>) + Send + Sync + 'static,
+        OA: Fn(SqsCompletionHandlerActor<CE, ProcErr>, Result<String, String>) + Send + Sync + 'static,
+        CacheT: Cache<CacheErr> + Send + Sync + Clone + 'static,
+        CacheErr: Debug + Clone + Send + Sync + 'static,
+        ProcErr: Debug + Clone + Send + Sync + 'static,
 {
     pub fn new(
         sqs_client: SqsClient,
@@ -73,33 +83,64 @@ impl<CPE, CP, CE, Payload, EE, OA> SqsCompletionHandler<CPE, CP, CE, Payload, EE
         event_emitter: EE,
         completion_policy: CompletionPolicy,
         on_ack: OA,
+        cache: CacheT,
     ) -> Self {
         Self {
             sqs_client,
             queue_url,
-            completed_events: Vec::with_capacity(10),
-            completed_messages: Vec::with_capacity(10),
+            completed_events: Vec::with_capacity(completion_policy.max_messages as usize),
+            identities: Vec::with_capacity(completion_policy.max_messages as usize),
+            completed_messages: Vec::with_capacity(completion_policy.max_messages as usize),
             completion_serializer,
             event_emitter,
             completion_policy,
             on_ack,
             self_actor: None,
+            cache,
+            _p: std::marker::PhantomData,
         }
     }
 }
 
-impl<CPE, CP, CE, Payload, EE, OA> SqsCompletionHandler<CPE, CP, CE, Payload, EE, OA>
+impl<CPE, CP, CE, Payload, EE, OA, CacheT, CacheErr, ProcErr> SqsCompletionHandler<CPE, CP, CE, Payload, EE, OA, CacheT, CacheErr, ProcErr>
     where
         CPE: Debug + Send + Sync + 'static,
         CP: CompletionEventSerializer<CompletedEvent=CE, Output=Payload, Error=CPE> + Send + Sync + 'static,
         Payload: Send + Sync + 'static,
         CE: Send + Sync + Clone + 'static,
         EE: EventEmitter<Event=Payload> + Send + Sync + 'static,
-        OA: Fn(SqsCompletionHandlerActor<CE>, Result<String, String>) + Send + Sync + 'static,
+        OA: Fn(SqsCompletionHandlerActor<CE, ProcErr>, Result<String, String>) + Send + Sync + 'static,
+        CacheT: Cache<CacheErr> + Send + Sync + Clone + 'static,
+        CacheErr: Debug + Clone + Send + Sync + 'static,
+        ProcErr: Debug + Clone + Send + Sync + 'static
 {
-    pub async fn mark_complete(&mut self, sqs_message: SqsMessage, completed: CE) {
-        self.completed_events.push(completed);
-        self.completed_messages.push(sqs_message);
+    pub async fn mark_complete(
+        &mut self,
+        sqs_message: SqsMessage,
+        completed: OutputEvent<CE, ProcErr>
+    ) {
+        match completed.completed_event {
+            Completion::Total(ce) => {
+                info!("Marking all events complete - total success");
+                self.completed_events.push(ce);
+                self.completed_messages.push(sqs_message);
+                self.identities.extend(completed.identities);
+            },
+            Completion::Partial((ce, err)) => {
+                warn!("EventHandler was only partially successful: {:?}", err);
+                self.completed_events.push(ce);
+                self.identities.extend(completed.identities);
+            },
+            Completion::Error(e) => {
+                warn!("Event handler failed: {:?}", e);
+            }
+        };
+
+        info!(
+            "Marked event complete. {} completed events, {} completed messages",
+            self.completed_events.len(),
+            self.completed_messages.len(),
+        );
 
         if self.completion_policy.should_flush(self.completed_events.len() as u16) {
             self.ack_all(None).await;
@@ -109,6 +150,7 @@ impl<CPE, CP, CE, Payload, EE, OA> SqsCompletionHandler<CPE, CP, CE, Payload, EE
 
     pub async fn ack_all(&mut self, notify: Option<tokio::sync::oneshot::Sender<()>>) {
         info!("Flushing completed events");
+
         let serialized_event = self
             .completion_serializer
             .serialize_completed_events(&self.completed_events[..]);
@@ -117,12 +159,10 @@ impl<CPE, CP, CE, Payload, EE, OA> SqsCompletionHandler<CPE, CP, CE, Payload, EE
             Ok(serialized_event) => serialized_event,
             Err(e) => {
                 // We should emit a failure, but ultimately we just have to not ack these messages
-                warn!("Serializing events failed: {:?}", e);
-
                 self.completed_events.clear();
                 self.completed_messages.clear();
 
-                return;
+                panic!("Serializing events failed: {:?}", e);
             }
         };
 
@@ -131,6 +171,12 @@ impl<CPE, CP, CE, Payload, EE, OA> SqsCompletionHandler<CPE, CP, CE, Payload, EE
             "Failed to emit event"
         );
 
+        for identity in self.identities.drain(..) {
+            if let Err(e) = self.cache.store(identity).await {
+                warn!("Failed to cache with: {:?}", e);
+            }
+        }
+
         let mut acks = vec![];
 
         for chunk in self.completed_messages.chunks(10) {
@@ -192,26 +238,30 @@ impl<CPE, CP, CE, Payload, EE, OA> SqsCompletionHandler<CPE, CP, CE, Payload, EE
 }
 
 #[allow(non_camel_case_types)]
-pub enum SqsCompletionHandlerMessage<CE>
+pub enum SqsCompletionHandlerMessage<CE, ProcErr>
     where
         CE: Send + Sync + Clone + 'static,
+        ProcErr : Debug + Send + Sync + Clone + 'static,
 {
-    mark_complete { msg: SqsMessage, completed: CE },
+    mark_complete { msg: SqsMessage, completed: OutputEvent<CE, ProcErr>  },
     ack_all { notify: Option<tokio::sync::oneshot::Sender<()>> },
 }
 
-impl<CPE, CP, CE, Payload, EE, OA> SqsCompletionHandler<CPE, CP, CE, Payload, EE, OA>
+impl<CPE, CP, CE, Payload, EE, OA, CacheT, CacheErr, ProcErr> SqsCompletionHandler<CPE, CP, CE, Payload, EE, OA, CacheT, CacheErr, ProcErr>
     where
         CPE: Debug + Send + Sync + 'static,
         CP: CompletionEventSerializer<CompletedEvent=CE, Output=Payload, Error=CPE> + Send + Sync + 'static,
         Payload: Send + Sync + 'static,
         CE: Send + Sync + Clone + 'static,
         EE: EventEmitter<Event=Payload> + Send + Sync + 'static,
-        OA: Fn(SqsCompletionHandlerActor<CE>, Result<String, String>) + Send + Sync + 'static,
+        OA: Fn(SqsCompletionHandlerActor<CE, ProcErr>, Result<String, String>) + Send + Sync + 'static,
+        CacheT: Cache<CacheErr> + Send + Sync + Clone + 'static,
+        CacheErr: Debug + Clone + Send + Sync + 'static,
+        ProcErr: Debug + Clone + Send + Sync + 'static,
 {
-    pub async fn route_message(&mut self, msg: SqsCompletionHandlerMessage<CE>) {
+    pub async fn route_message(&mut self, msg: SqsCompletionHandlerMessage<CE, ProcErr>) {
         match msg {
-            SqsCompletionHandlerMessage::mark_complete { msg, completed } => {
+            SqsCompletionHandlerMessage::mark_complete { msg, completed} => {
                 self.mark_complete(msg, completed).await
             }
             SqsCompletionHandlerMessage::ack_all { notify } => self.ack_all(notify).await,
@@ -220,18 +270,20 @@ impl<CPE, CP, CE, Payload, EE, OA> SqsCompletionHandler<CPE, CP, CE, Payload, EE
 }
 
 #[derive(Clone)]
-pub struct SqsCompletionHandlerActor<CE>
+pub struct SqsCompletionHandlerActor<CE, ProcErr>
     where
         CE: Send + Sync + Clone + 'static,
+        ProcErr:  Debug + Send + Sync + Clone + 'static,
 {
-    sender: Sender<SqsCompletionHandlerMessage<CE>>,
+    sender: Sender<SqsCompletionHandlerMessage<CE, ProcErr>>,
 }
 
-impl<CE> SqsCompletionHandlerActor<CE>
+impl<CE, ProcErr> SqsCompletionHandlerActor<CE, ProcErr>
     where
         CE: Send + Sync + Clone + 'static,
+        ProcErr:  Debug + Send + Sync + Clone + 'static,
 {
-    pub fn new<CPE, CP, Payload, EE, OA>(mut actor_impl: SqsCompletionHandler<CPE, CP, CE, Payload, EE, OA>) -> Self
+    pub fn new<CPE, CP, Payload, EE, OA, CacheT, CacheErr>(mut actor_impl: SqsCompletionHandler<CPE, CP, CE, Payload, EE, OA, CacheT, CacheErr, ProcErr>) -> Self
         where
             CPE: Debug + Send + Sync + 'static,
             CP: CompletionEventSerializer<CompletedEvent=CE, Output=Payload, Error=CPE>
@@ -241,7 +293,10 @@ impl<CE> SqsCompletionHandlerActor<CE>
             + 'static,
             Payload: Send + Sync + 'static,
             EE: EventEmitter<Event=Payload> + Send + Sync + 'static,
-            OA: Fn(SqsCompletionHandlerActor<CE>, Result<String, String>) + Send + Sync + 'static,
+            OA: Fn(SqsCompletionHandlerActor<CE, ProcErr>, Result<String, String>) + Send + Sync + 'static,
+            CacheT: Cache<CacheErr> + Send + Sync + Clone + 'static,
+            CacheErr: Debug + Clone + Send + Sync + 'static,
+
     {
         let (sender, receiver) = channel(1);
 
@@ -259,8 +314,8 @@ impl<CE> SqsCompletionHandlerActor<CE>
         self_actor
     }
 
-    pub async fn mark_complete(&self, msg: SqsMessage, completed: CE) {
-        let msg = SqsCompletionHandlerMessage::mark_complete { msg, completed };
+    pub async fn mark_complete(&self, msg: SqsMessage, completed: OutputEvent<CE, ProcErr>) {
+        let msg = SqsCompletionHandlerMessage::mark_complete { msg, completed};
         if let Err(_e) = self.sender.clone().send(msg).await {
             panic!("Receiver has failed, propagating error. mark_complete")
         }
@@ -283,16 +338,17 @@ pub trait CompletionHandler {
 }
 
 #[async_trait]
-impl<CE> CompletionHandler for SqsCompletionHandlerActor<CE>
+impl<CE, ProcErr> CompletionHandler for SqsCompletionHandlerActor<CE, ProcErr>
     where
         CE: Send + Sync + Clone + 'static,
+        ProcErr:  Debug + Send + Sync + Clone + 'static,
 {
     type Message = SqsMessage;
-    type CompletedEvent = CE;
+    type CompletedEvent = OutputEvent<CE, ProcErr>;
 
     async fn mark_complete(&self, msg: Self::Message, completed_event: Self::CompletedEvent) {
         SqsCompletionHandlerActor::mark_complete(
-            self, msg, completed_event,
+            self, msg, completed_event
         ).await
     }
 
@@ -303,28 +359,34 @@ impl<CE> CompletionHandler for SqsCompletionHandlerActor<CE>
     }
 }
 
-pub struct SqsCompletionHandlerRouter<CPE, CP, CE, Payload, EE, OA>
+pub struct SqsCompletionHandlerRouter<CPE, CP, CE, Payload, EE, OA, CacheT, CacheErr, ProcErr>
     where
         CPE: Debug + Send + Sync + 'static,
         CP: CompletionEventSerializer<CompletedEvent=CE, Output=Payload, Error=CPE> + Send + Sync + 'static,
         Payload: Send + Sync + 'static,
         CE: Send + Sync + Clone + 'static,
         EE: EventEmitter<Event=Payload> + Send + Sync + 'static,
-        OA: Fn(SqsCompletionHandlerActor<CE>, Result<String, String>) + Send + Sync + 'static,
+        OA: Fn(SqsCompletionHandlerActor<CE, ProcErr>, Result<String, String>) + Send + Sync + 'static,
+        CacheT: Cache<CacheErr> + Send + Sync + Clone + 'static,
+        CacheErr: Debug + Clone + Send + Sync + 'static,
+        ProcErr: Debug + Clone + Send + Sync + 'static,
 {
-    receiver: Receiver<SqsCompletionHandlerMessage<CE>>,
-    actor_impl: SqsCompletionHandler<CPE, CP, CE, Payload, EE, OA>,
+    receiver: Receiver<SqsCompletionHandlerMessage<CE, ProcErr>>,
+    actor_impl: SqsCompletionHandler<CPE, CP, CE, Payload, EE, OA, CacheT, CacheErr, ProcErr>,
 }
 
 
-async fn route_wrapper<CPE, CP, CE, Payload, EE, OA>(mut router: SqsCompletionHandlerRouter<CPE, CP, CE, Payload, EE, OA>)
+async fn route_wrapper<CPE, CP, CE, Payload, EE, OA, CacheT, CacheErr, ProcErr>(mut router: SqsCompletionHandlerRouter<CPE, CP, CE, Payload, EE, OA, CacheT, CacheErr, ProcErr>)
     where
         CPE: Debug + Send + Sync + 'static,
         CP: CompletionEventSerializer<CompletedEvent=CE, Output=Payload, Error=CPE> + Send + Sync + 'static,
         Payload: Send + Sync + 'static,
         CE: Send + Sync + Clone + 'static,
         EE: EventEmitter<Event=Payload> + Send + Sync + 'static,
-        OA: Fn(SqsCompletionHandlerActor<CE>, Result<String, String>) + Send + Sync + 'static,
+        OA: Fn(SqsCompletionHandlerActor<CE, ProcErr>, Result<String, String>) + Send + Sync + 'static,
+        CacheT: Cache<CacheErr> + Send + Sync + Clone + 'static,
+        CacheErr: Debug + Clone + Send + Sync + 'static,
+        ProcErr: Debug + Clone + Send + Sync + 'static,
 {
     while let Some(msg) = router.receiver.recv().await {
         router.actor_impl.route_message(msg).await;
